#ifndef _ASM_ALPHA_PARAM_H
#define _ASM_ALPHA_PARAM_H

#ifndef HZ
# define HZ	1024
#endif

#define EXEC_PAGESIZE	8192

#ifndef NGROUPS
#define NGROUPS		32
#endif

#ifndef NOGROUP
#define NOGROUP		(-1)
#endif

#define MAXHOSTNAMELEN	64	/* max length of hostname */

#endif /* _ASM_ALPHA_PARAM_H */
