/*
 * linux/include/asm-arm/arch-nexuspci/dma.h
 *
 * Architecture DMA routines - NexusPCI has no DMA, so this file is very
 * simple. :-)
 *
 * Copyright (C) 1998 Philip Blundell
 */

/* Need this to keep <asm/dma.h> happy. */
typedef unsigned int dmamode_t;

